/* ©  Copyright Steve Simon, 2013 */
#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <auth.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include <bio.h>
#include <json.h>
#include "jira.h"

Node *
uservalid(Node *n, Sess *s, char *user)
{
	Node *cc;
	Jvalue *v;

	if((v = exch(s, nil, "rest/api/%s/user?username=%s&", s->api, user)) == nil)
		return nil;
	json_free(v);
	cc = subdir(&n->child, user);

	scanhook(newdir(&cc, user, nil, user), userinfo, user);

	return cc;
}


Node *
userinfo(Node *n, Sess *s, char *user)
{
	int i;
	Fmt f;
	Jvalue *v;
	char *p, *buf;

	if((v = exch(s, nil, "rest/api/%s/user?username=%s&expand=groups", s->api, user)) == nil){
		warn("%s cannot get user detail, %r\n", user);
		return n;
	}
	if((p = json_strval(v, '.', "displayName")) != nil)
		subtext(&n, "realname", p);
	if((p = json_strval(v, '.', "emailAddress")) != nil)
		subtext(&n, "email", p);
	if((p = json_strval(v, '.', "active")) != nil)
		subtext(&n, "active", p);
	if((p = json_strval(v, '.', "timeZone")) != nil)
		subtext(&n, "timezone", p);

	fmtstrinit(&f);
	for(i = 0; (p = json_strval(v, '.', "groups.items[%d].name", i)) != nil; i++)
		fmtprint(&f, "%s\n", p);
	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	subtext(&n, "groups", buf);
	free(buf);

	scanhook(newdir(&n, nil, nil, "queue"), queuelist, user);
	scanhook(newdir(&n, nil, nil, "reported"), reportedlist, user);
	scanhook(newdir(&n, nil, nil, "resolved"), resolvedlist, user);

	json_free(v);
	return n;
}

/*
 * The json_ira rest api has no way to enumerate the usernames known to jira
 * so we fake it by adding a /users dirrectory each username as we find
 * them.
 */
void
adduser(Sess *s, char *uid)
{
	Node *n;

	for(n = *(s->users); n; n = n->next)
		if(strcmp(n->dir.name, uid) == 0)
			return;
	scanhook(newdir(s->users, uid, nil, uid), userinfo, uid);
}

Node *
issuevalid(Node *n, Sess *s, char *issue)
{
	Node *cc;
	Jvalue *v;
	char *lwr;

	if((v = exch(s, nil, "rest/api/%s/issue/%U/transitions", s->api, issue)) == nil)
		return nil;
	json_free(v);

	cc = subdir(&n->child, issue);
	lwr = strlwr(estrdup9p(issue));
	scanhook(newdir(&cc, lwr, nil, lwr), s->issueinfo, issue);
	free(lwr);

	return cc;
}

Node *
issuelist(Node *n, Sess *s, char *proj)		/* v5 compat */
{
	Jvalue *v;
	int i, want, start;
	char *fmt, *issue, *lwr;

	start = 0;
	want = 1000;
	fmt = "rest/api/%s/search?jql=project=%U&maxResults=%d&startAt=%d";
	do{
		if((v = exch(s, nil, fmt, s->api, proj, want, start)) == nil){
			warn("%s cannot get issue list, %r\n", proj);
			break;
		}
		for(i = 0; (issue = json_strval(v, '.', "issues[%d].key", i)) != nil; i++){
			lwr = strlwr(issue);
			scanhook(newdir(&n, lwr, nil, lwr), s->issueinfo, issue);
		}
		json_free(v);
		start += i;
	}while(i == want);
	return n;
}

static void
issuetypes(Node **np, Jvalue *v)
{
	int i;
	Fmt f;
	char *p, *buf;

	fmtstrinit(&f);

	fmtprint(&f, "%-24q %-5s %-12s %s\n", "#Name", "Id", "Subtask", "Desc");
	for(i = 0; (p = json_strval(v, '.', "issueTypes[%d].name", i)) != nil; i++){
		fmtprint(&f, "%-24q ", p);
		if((p = json_strval(v, '.', "issueTypes[%d].id", i)) != nil)
			fmtprint(&f, "%-5q", p);
		else
			fmtprint(&f, "%-5s", "-");
		if((p = json_strval(v, '.', "issueTypes[%d].subtask", i)) != nil)
			fmtprint(&f, "%-12q", p);
		else
			fmtprint(&f, "%-12s", "-");
		if((p = json_strval(v, '.', "issueTypes[%d].description", i)) != nil)
			fmtprint(&f, "%s", p);
		fmtprint(&f, "\n");
	}
	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	subtext(np, "issuetypes", buf);
	free(buf);
}

static void
versions(Node **np, Jvalue *v)		/* FIXME: Not checked for v5 compat yet */
{
	int i;
	Fmt f;
	char *p, *buf;

	fmtstrinit(&f);
	fmtprint(&f, "%-16s %-10s %-10s %s\n", "#Name", "Archived", "Released", "Description");
	for(i = 0; (p = json_strval(v, '.', "versions[%d].name", i)) != nil; i++){
		fmtprint(&f, "%-16q ", p);

		if((p = json_strval(v, '.', "versions[%d].archived", i)) != nil)
			fmtprint(&f, "%-10q ", p);
		else
			fmtprint(&f, "%-10s ", "-");
		if((p = json_strval(v, '.', "versions[%d].released", i)) != nil)
			fmtprint(&f, "%-10q ", p);
		else
			fmtprint(&f, "%-10s ", "-");
		if((p = json_strval(v, '.', "versions[%d].description", i)) != nil)
			fmtprint(&f, "%s\n", p);
		else
			fmtprint(&f, "%s\n", "-");
	}
	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(np, "versions", buf);
	free(buf);
}


static void
components(Node *n, Jvalue *v)
{
	int i;
	Node *c;
	char *p, *comp;

	for(i = 0; (comp = json_strval(v, '.', "components[%d].name", i)) != nil; i++){
		c = newdir(&n->child, nil, nil, comp);
		if((p = json_strval(v, '.', "components[%d].description", i)) != nil)
			subtext(&c->child, "description", p);
		if((p = json_strval(v, '.', "components[%d].lead.name", i)) != nil)
			subtext(&c->child, "lead", p);
		if((p = json_strval(v, '.', "components[%d].assignee.name", i)) != nil)
			subtext(&c->child, "assignee", p);
		if((p = json_strval(v, '.', "components[%d].realAssignee.name", i)) != nil)
			subtext(&c->child, "realassignee", p);
	}
}


Node *
projvalid(Node *n, Sess *s, char *proj)
{
	Node *cc, *cc2;
	Jvalue *v;
	char *lwr;

	if((v = exch(s, nil, "rest/api/%s/project/%U/versions", s->api, proj)) == nil)
		return nil;
	json_free(v);

	cc = subdir(&n->child, proj);

	lwr = strlwr(estrdup9p(proj));
	cc2 = newdir(&cc, lwr, nil, lwr);
	scanhook(cc2, projinfo, proj);

	cc2 = subdir(&cc->child, "issues");		/* Temporary bodge until "walk not exist, use scan for walk" code is done */
	scanhook(cc2, issuelist, proj);
	walkhook(cc2, issuevalid, proj);

	free(lwr);
	return cc;
}

Node *
assignable(Node *n, Sess *s, char *proj)
{
	Fmt f;
	Jvalue *v;
	char *p, *fmt, *buf;
	int i, start, want;


	/*
	 * this only works if the current user has permission to
	 * perform issue reassignment in this project, this is to
	 * prevent email address harvesting by spammers.
	 */
	start = 0;
//	want = 1000;
want = 10;				// FIXME: bodge for debug
	fmtstrinit(&f);
	fmt = "/rest/api/%s/user/assignable/search?project=%U&maxResults=%d&startAt=%d";
	do{
		if((v = exch(s, nil, fmt, s->api, proj, want, start)) == nil){
			warn("%s cannot get assignable user list, %r\n", proj);
			return n;
		}

/*
 * never seen one of these in the wild
 * so I don't know how to decode it,
 * this is guesswork!
 */
		for(i = 0; i < want && (p = json_strval(v, '.', "assignable[%d].name", i)) != nil; i++)
			fmtprint(&f, "%s\n", p);
		json_free(v);
		start += i;
			
		json_free(v);
		start += i;
	}while(i == want);

	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(&n, "watchers", buf);
	free(buf);
	return n;
}

Node *
projinfo(Node *n, Sess *s, char *proj)
{
	Node *cc;
	Jvalue *v;
	char *p, *uid;

	if((v = exch(s, nil, "rest/api/%s/project/%U", s->api, proj)) == nil){
		warn("%U cannot get project detail, %r\n", proj);
		return n;
	}

	if((uid = json_strval(v, '.', "lead.name")) != nil){
		fulltext(&n, "lead", uid, nil, uid, 0, 0, uid);
		adduser(s, uid);
	}
	if((p = json_strval(v, '.', "name")) != nil)
		fulltext(&n, "title", uid, nil, uid, 0, 0, p);
	if((p = json_strval(v, '.', "description")) != nil)
		fulltext(&n, "description", uid, nil, uid, 0, 0, p);

	issuetypes(&n, v);
	versions(&n, v);

	components(newdir(&n, uid, nil, "components"), v);

	cc = newdir(&n, uid, nil, "issues");
	scanhook(cc, issuelist, proj);
	walkhook(cc, issuevalid, proj);

	if(s->major >= 5)
		scanhook(newdir(&n, uid, nil, "assignable"), assignable, proj);

	json_free(v);
	return n;
}

Node *
queuelist(Node *n, Sess *s, char *user)
{
	Jvalue *v;
	int i, start, want;
	char *issue, *fmt, *lwr;

	start = 0;
	want = 1000;
	fmt = "rest/api/%s/search?jql=assignee=%s+and+status!=Resolved&maxResults=%d&startAt=%d";
	do{
		if((v = exch(s, nil, fmt, s->api, user, want, start)) == nil){
			warn("%s cannot get user's queue, %r\n", user);
			break;
		}
		for(i = 0; (issue = json_strval(v, '.', "issues[%d].key", i)) != nil; i++){
			lwr = strlwr(estrdup9p(issue));
			scanhook(newdir(&n, nil, nil, lwr), s->issueinfo, issue);
			free(lwr);
		}
		json_free(v);
		start += i;
	}while(i == want);
	return n;
}

Node *
resolvedlist(Node *n, Sess *s, char *user)
{
	Jvalue *v;
	int i, start, want;
	char *issue, *fmt;

	start = 0;
	want = 1000;
	fmt = "rest/api/%s/search?jql=assignee=%s+and+status=Resolved&maxResults=%d&startAt=%d";
	do{
		if((v = exch(s, nil, fmt, s->api, user, want, start)) == nil){
			warn("%s cannot get user's queue, %r\n", user);
			break;
		}
		for(i = 0; (issue = json_strval(v, '.', "issues[%d].key", i)) != nil; i++)
			scanhook(newdir(&n, nil, nil, strlwr(issue)), s->issueinfo, issue);
		json_free(v);
		start += i;
	}while(i == want);
	return n;
}

Node *
reportedlist(Node *n, Sess *s, char *user)
{
	Jvalue *v;
	int i, start, want;
	char *issue, *fmt, *lwr;

	start = 0;
	want = 1000;
	fmt = "rest/api/%s/search?jql=reporter=%s&maxResults=%d&startAt=%d";
	do{
		if((v = exch(s, nil, fmt, s->api, user, want, start)) == nil){
			warn("%s cannot get user's queue, %r\n", user);
			break;
		}
		for(i = 0; (issue = json_strval(v, '.', "issues[%d].key", i)) != nil; i++){
			lwr = strlwr(issue);
			scanhook(newdir(&n, nil, nil, lwr), s->issueinfo, issue);
			free(lwr);
		}
		json_free(v);
		start += i;
	}while(i == want);
	return n;
}



Node *
projlist(Node *n, Sess *s, char *)
{
	int i;
	Jvalue *v;
	char *proj, *lwr;

	if((v = exch(s, nil, "rest/api/%s/project", s->api)) == nil){
		warn("cannot get project list, %r\n");
		return n;
	}

	for(i = 0; (proj = json_strval(v, '.', "[%d].key", i)) != nil; i++){
		lwr = strlwr(proj);
		scanhook(newdir(&n, lwr, nil, lwr), projinfo, proj);
	}
	json_free(v);
	return n;
}

Node *
serverinfo(Node *n, Sess *s, char *)
{
	char *p;
	Jvalue *v;

	if((v = exch(s, nil, "rest/api/%s/serverInfo", s->api)) == nil){
		warn("cannot get server detail, %r\n");
		return n;
	}


	if((p = json_strval(v, '.', "version")) != nil){
		subtext(&n, "version", p);
		s->major = strtol(p, nil, 10);
	}
	if((p = json_strval(v, '.', "buildNumber")) != nil)
		subtext(&n, "build-number", p);
	if((p = json_strval(v, '.', "buildDate")) != nil)
		subtext(&n, "build-time", ctime(iso2sec(p)));
	if((p = json_strval(v, '.', "buildPartnerName")) != nil)
		subtext(&n, "build-partner", p);
	if((p = json_strval(v, '.', "serverTime")) != nil)
		subtext(&n, "server-time", ctime(iso2sec(p)));
	if((p = json_strval(v, '.', "serverTitle")) != nil)
		subtext(&n, "title", p);
	if((p = json_strval(v, '.', "scmInfo")) != nil)
		subtext(&n, "scm-info", p);
	if((p = json_strval(v, '.', "baseUrl")) != nil)
		subtext(&n, "base-url", p);
	else
		subtext(&n, "base-url", s->url);
	json_free(v);
	return n;
}

Node *
sessioninfo(Node *n, Sess *s, char *)
{
	char *p;
	Jvalue *v;

	/* This may be overwritten depending on server setup */
	if(s->annon)
		subtext(&n, "username", "<annonymous>");

	if((v = exch(s, nil, "rest/auth/1/session")) == nil){
		warn("cannot get session detail, %r\n");
		return n;
	}
	if((p = json_strval(v, '.', "name")) != nil){
		adduser(s, p);
		free(s->user);
		s->user = estrdup9p(p);
		fulltext(&n, "username", s->user, nil, s->user, 0, 0, s->user);
	}
	if((p = json_strval(v, '.', "loginInfo.loginCount")) != nil)
		fulltext(&n, "logins", s->user, nil, s->user, 0, 0, p);
	if((p = json_strval(v, '.', "loginInfo.previousLoginTime")) != nil)
		fulltext(&n, "last-login", s->user, nil, s->user, 0, 0, ctime(iso2sec(p)));
	if((p = json_strval(v, '.', "loginInfo.failedLoginCount")) != nil)
		fulltext(&n, "failed-logins", s->user, nil, s->user, 0, 0, p);
	if((p = json_strval(v, '.', "loginInfo.lastFailedLoginTime")) != nil)
		fulltext(&n, "last-failed-login", s->user, nil, s->user, 0, 0, ctime(iso2sec(p)));
	json_free(v);
	return n;
}

