/* ©  Copyright Steve Simon, 2013 */
#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <auth.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include <bio.h>
#include <json.h>
#include "jira.h"


static void
links(Node *n, Jvalue *v)
{
	int i;
	char *p;
	Node *c;

	for(i = 0; (p = json_strval(v, '.', "fields.links.value[%d].issueKey", i)) != nil; i++){
		c = newdir(&n->child, nil, nil, p);
		if((p = json_strval(v, '.', "fields.tasks.value[%d].type.name", i)) != nil)
			subtext(&c->child, "name", p);
		if((p = json_strval(v, '.', "fields.tasks.value[%d].type.direction", i)) != nil)
			subtext(&c->child, "direction", p);
		if((p = json_strval(v, '.', "fields.tasks.value[%d].type.description", i)) != nil)
			subtext(&c->child, "description", p);
	}
}

static void
subtasks(Node *n, Jvalue *v)
{
	int i;
	char *p;
	Node *c;

	for(i = 0; (p = json_strval(v, '.', "fields.sub-tasks.value[%d].key", i)) != nil; i++){
		c = newdir(&n->child, nil, nil, p);
		if((p = json_strval(v, '.', "fields.tasks.value[%d].type.name", i)) != nil)
			subtext(&c->child, "name", p);
		if((p = json_strval(v, '.', "fields.tasks.value[%d].type.direction", i)) != nil)
			subtext(&c->child, "direction", p);
	}
}


static void
attachments(Node *n, Jvalue *v, Sess *s)
{
	int i;
	long created;
	vlong length;
	char *p, *author, *file, *content;

	for(i = 0; (file = json_strval(v, '.', "fields.attachment.value[%d].filename", i)) != nil; i++){

		if((author = json_strval(v, '.', "fields.attachment.value[%d].author.name", i)) != nil)
			adduser(s, author);
		else
			author = "unknown";

		if((p = json_strval(v, '.', "fields.attachment.value[%d].created", i)) != nil)
			created = iso2sec(p);
		else
			created = time(nil);

		if((p = json_strval(v, '.', "fields.attachment.value[%d].size", i)) != nil)
			length = strtoull(p, nil, 10);
		else
			length = 0;

		if((content = json_strval(v, '.', "fields.attachment.value[%d].content", i)) != nil)
			fetchtext(&n->child, file, author, nil, author, created, length, content);
	}
}

static void
worklog(Node *n, Jvalue *v, Sess *s)
{
	int i;
	Node *c;
	char *p, dir[16];

	for(i = 0; (p = json_strval(v, '.', "fields.worklog.value[%d].author.name", i)) != nil; i++){
		snprint(dir, sizeof(dir), "%d", i);
		c = newdir(&n->child, dir, nil, p);

		subtext(&c->child, "author", p);
		adduser(s, p);

		if((p = json_strval(v, '.', "fields.worklog.value[%d].updateAuthor.name", i)) != nil)
			subtext(&c->child, "updater", p);
		adduser(s, p);

		if((p = json_strval(v, '.', "fields.worklog.value[%d].comment", i)) != nil)
			subtext(&c->child, "comment", p);
		if((p = json_strval(v, '.', "fields.worklog.value[%d].started", i)) != nil)
			subtext(&c->child, "started", ctime(iso2sec(p)));
		if((p = json_strval(v, '.', "fields.worklog.value[%d].minutesSpent", i)) != nil)
			subtext(&c->child, "minuites", p);
	}
}

static void
timetracking(Node **np, Jvalue *v)
{
	Fmt f;
	char *p, *buf;

	fmtstrinit(&f);
	if((p = json_strval(v, '.', "fields.timetracking.value.timeoriginalestimate")) != nil)
		fmtprint(&f, "%-16q %q\n", "originalestimate", p);
	if((p = json_strval(v, '.', "fields.timetracking.value.timeestimate")) != nil)
		fmtprint(&f, "%-16q %q\n", "estimate", p);
	if((p = json_strval(v, '.', "fields.timetracking.value.timespent")) != nil)
		fmtprint(&f, "%-16q %q\n", "spent", p);
	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(np, "timetracking", buf);
	free(buf);
}

static void
transitions(Node **n, Sess *s, char *issue)
{
	int i;
	Fmt f;
	Jvalue *v;
	char *buf, *name, *id;

	if((v = exch(s, nil, "rest/api/%s/issue/%U/transitions", s->api, issue)) == nil){
		warn("%s cannot get valid transition list, %r\n", issue);
		return;
	}

	fmtstrinit(&f);
	for(i = 0; i < 255; i++){
		if((id = json_strval(v, '.', "%d.transitionDestination", i)) == nil)
			continue;
		if((name = json_strval(v, '.', "%d.name", i)) == nil)
			name = "";
		fmtprint(&f, "%-4q %q\n", id, name);
	}
	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	writehook(subtext(n, "transition", buf), add_transition, issue);
	free(buf);
	json_free(v);
}


static void
comments(Node *n, Jvalue *v, Sess *s)
{
	int i;
	long created, updated;
	char *p, *author, *body, file[16];

	for(i = 0; (body = json_strval(v, '.', "fields.comment.value[%d].body", i)) != nil; i++){

		if((author = json_strval(v, '.', "fields.comment.value[%d].author.name", i)) == nil)
			author = "unknown";
		else
			adduser(s, author);

		if((p = json_strval(v, '.', "fields.comment.value[%d].created", i)) != nil)
			created = iso2sec(p);
		else
			created = time(nil);

		if((p = json_strval(v, '.', "fields.comment.value[%d].updated", i)) != nil)
			updated = iso2sec(p);
		else
			updated = time(nil);

		snprint(file, sizeof(file), "%d", i);
		fulltext(&n->child, file, author, nil, author, created, updated, body);
	}
}

Node *
issueinfo_v4(Node *n, Sess *s, char *issue)
{
	Jvalue  *v;
	char *p, *uid;

	if((v = exch(s, nil, "rest/api/%s/issue/%U", s->api, issue)) == nil){
		warn("%s cannot get issue detail, %r\n", issue);
		return n;
	}

	if((uid = json_strval(v, '.', "fields.reporter.value.name")) != nil){
		adduser(s, uid);
		fulltext(&n, "reporter", uid, nil, uid, 0, 0, uid);
	}
	if((p = json_strval(v, '.', "fields.status.value.name")) != nil)
		fulltext(&n, "status", uid, nil, uid, 0, 0, p);
	if((p = json_strval(v, '.', "fields.resolution.value.name")) != nil)
		fulltext(&n, "resolution", uid, nil, uid, 0, 0, p);
	if((p = json_strval(v, '.', "fields.created")) != nil)
		fulltext(&n, "created", uid, nil, uid, 0, 0, ctime(iso2sec(p)));
	if((p = json_strval(v, '.', "fields.updated")) != nil)
		fulltext(&n, "updated", uid, nil, uid, 0, 0, ctime(iso2sec(p)));
	if((p = json_strval(v, '.', "fields.resolutiondate.value")) != nil)
		fulltext(&n, "resolved", uid, nil, uid, 0, 0, ctime(iso2sec(p)));
	if((p = json_strval(v, '.', "fields.environment.value")) != nil)
		fulltext(&n, "environment", uid, nil, uid, 0, 0, p);
	if((p = json_strval(v, '.', "FixVersions.value.name")) != nil)
		fulltext(&n, "fix-version", uid, nil, uid, 0, 0, p);

	if((p = json_strval(v, '.', "fields.assignee.value.name")) == nil)
		p = "";
	else
		adduser(s, p);
	writehook(fulltext(&n, "assignee", uid, nil, uid, 0, 0, p), upd_assignee, issue);

	if((p = json_strval(v, '.', "fields.issuetype.value.name")) == nil)
		p = "";
	writehook(fulltext(&n, "issuetype", uid, nil, uid, 0, 0, p), upd_summary, issue);

	if((p = json_strval(v, '.', "fields.summary.value")) == nil)
		p = "";
	writehook(fulltext(&n, "summary", uid, nil, uid, 0, 0, p), upd_summary, issue);

	if((p = json_strval(v, '.', "fields.description.value")) == nil)
		p = "";
	writehook(fulltext(&n, "description", uid, nil, uid, 0, 0, p), upd_description, issue);

	timetracking(&n, v);
	transitions(&n, s, issue);
	
	worklog(newdir(&n, nil, nil, "worklog"), v, s);
	comments(newdir(&n, nil, nil, "comments"), v, s);
	attachments(newdir(&n, nil, nil, "attachments"), v, s);
	subtasks(newdir(&n, nil, nil, "subtasks"), v);
	links(newdir(&n, nil, nil, "links"), v);

	json_free(v);
	return n;
}

