/* ©  Copyright Steve Simon, 2013 */
#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <auth.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include <bio.h>
#include <json.h>
#include "jira.h"

static void
links(Node *n, Jvalue *v)	// FIXME: not right at all...
{
	int i;
	char *p;

	for(i = 0; (p = json_strval(v, '.', "fields.issuelinks[%d].type.name", i)) != nil; i++){
			subtext(&n->child, "name", p);
		if((p = json_strval(v, '.', "fields.issuelinks[%d].outward", i)) != nil)
			subtext(&n->child, "outward", p);
		if((p = json_strval(v, '.', "fields.issuelinks[%d].type.direction", i)) != nil)
			subtext(&n->child, "direction", p);
		if((p = json_strval(v, '.', "fields.issuelinks[%d].type.description", i)) != nil)
			subtext(&n->child, "description", p);
	}
 }

static void
subtasks(Node *n, Jvalue *v)
{
	int i;
	char *p;
	Node *c;

	for(i = 0; (p = json_strval(v, '.', "fields.sub-tasks[%d].key", i)) != nil; i++){
		c = newdir(&n->child, nil, nil, p);
		if((p = json_strval(v, '.', "fields.tasks[%d].type.name", i)) != nil)
			subtext(&c->child, "name", p);
		if((p = json_strval(v, '.', "fields.tasks[%d].type.direction", i)) != nil)
			subtext(&c->child, "direction", p);
	}
}


static void
attachments(Node *n, Jvalue *v, Sess *s)
{
	int i;
	vlong length;
	long created;
	char *p, *author, *file, *content;

	for(i = 0; (file = json_strval(v, '.', "fields.attachment[%d].filename", i)) != nil; i++){

		if((author = json_strval(v, '.', "fields.attachment[%d].author.name", i)) != nil)
			adduser(s, author);
		else
			author = "unknown";

		if((p = json_strval(v, '.', "fields.attachment[%d].created", i)) != nil)
			created = iso2sec(p);
		else
			created = time(nil);

		if((p = json_strval(v, '.', "fields.attachment[%d].size", i)) != nil)
			length = strtoull(p, nil, 10);
		else
			length = 0;

		if((content = json_strval(v, '.', "fields.attachment[%d].content", i)) != nil)
			fetchtext(&n->child, file, author, nil, author, created, length, content);
	}
}

static void
watchers(Node **np, Jvalue *v)
{
	int i;
	Fmt f;
	char *p, *buf;

	fmtstrinit(&f);
	for(i = 0; (p = json_strval(v, '.', "watcher.watchers[%d].name", i)) != nil; i++)
		fmtprint(&f, "%s\n", p);
	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(np, "watchers", buf);
	free(buf);
}

static void
fixversions(Node **np, Jvalue *v)
{
	int i;
	Fmt f;
	char *p, *buf;

	fmtstrinit(&f);
	fmtprint(&f, "%-16s %-10s %-10s %-12s %s\n", "#Name", "Archived", "Released", "Rel-date", "Description");
	for(i = 0; (p = json_strval(v, '.', "fields.fixVersions[%d].name", i)) != nil; i++){
		fmtprint(&f, "%-16q ", p);

		if((p = json_strval(v, '.', "fields.fixVersions[%d].archived", i)) != nil)
			fmtprint(&f, "%-10q ", p);
		else
			fmtprint(&f, "%-10s ", "-");
		if((p = json_strval(v, '.', "fields.fixVersions[%d].released", i)) != nil)
			fmtprint(&f, "%-10q ", p);
		else
			fmtprint(&f, "%-10s ", "-");
		if((p = json_strval(v, '.', "fields.fixVersions[%d].releaseDate", i)) != nil)
			fmtprint(&f, "%-12q ", p);
		else
			fmtprint(&f, "%-12s ", "-");
		if((p = json_strval(v, '.', "fields.fixVersions[%d].description", i)) != nil)
			fmtprint(&f, "%s\n", p);
		else
			fmtprint(&f, "%s\n", "-");
	}
	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(np, "fixversions", buf);
	free(buf);
}

static void
components(Node **np, Jvalue *v)
{
	int i;
	Fmt f;
	char *p, *buf;

	fmtstrinit(&f);
	for(i = 0; (p = json_strval(v, '.', "fields.components[%d].name", i)) != nil; i++)
		fmtprint(&f, "%s\n", p);
	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(np, "components", buf);
	free(buf);
}

static void
progress(Node **np, Jvalue *v, char *path, char *file)
{
	Fmt f;
	char *p, *buf;

	fmtstrinit(&f);
	if((p = json_strval(v, '.', "%s.progress", path)) != nil)
		fmtprint(&f, "progress	%s\n", p);
	if((p = json_strval(v, '.', "%s.total", path)) != nil)
		fmtprint(&f, "total	%s\n", p);
	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(np, file, buf);
	free(buf);
}

static void
worklog(Node *n, Jvalue *v, Sess *s)
{
	int i;
	Node *c;
	char *p, *q, dir[16], buf[128];

	for(i = 0; (p = json_strval(v, '.', "fields.worklog[%d].author.name", i)) != nil; i++){
		snprint(dir, sizeof(dir), "%d", i);
		c = newdir(&n->child, dir, nil, p);

		subtext(&c->child, "author", p);
		adduser(s, p);

		if((p = json_strval(v, '.', "fields.worklog[%d].updateAuthor.name", i)) != nil)
			subtext(&c->child, "updater", p);
		adduser(s, p);

		if((p = json_strval(v, '.', "fields.worklog[%d].comment", i)) != nil)
			subtext(&c->child, "comment", p);
		if((p = json_strval(v, '.', "fields.worklog[%d].started", i)) != nil)
			subtext(&c->child, "started", ctime(iso2sec(p)));
		if((p = json_strval(v, '.', "fields.worklog[%d].timeSpent", i)) != nil){
			if((q = json_strval(v, '.', "fields.worklog[%d].timeSpentSeconds", i)) != nil)
				snprint(buf, sizeof(buf), "%s %s", p, q);
			else
				snprint(buf, sizeof(buf), "%s", p);
			subtext(&c->child, "minuites", p);
		}
	}
}

static void
timetracking(Node **np, Jvalue *v)
{
	Fmt f;
	char *p, *q, *buf;

	fmtstrinit(&f);
	if((p = json_strval(v, '.', "fields.timetracking.timeoriginalestimate")) != nil)
		if((q = json_strval(v, '.', "fields.timetracking.originalEstimateSeconds")) != nil)
			fmtprint(&f, "%-16q %q %q\n", "originalEstimate", p, q);
		else
			fmtprint(&f, "%-16q %q\n", "originalEstimate", p);

	if((p = json_strval(v, '.', "fields.timetracking.remainingEstimate")) != nil)
		if((q = json_strval(v, '.', "fields.timetracking.remainingEstimateSeconds")) != nil)
			fmtprint(&f, "%-16q %q %q\n", "remainingEstimate", p, q);
		else
			fmtprint(&f, "%-16q %q\n", "remainingEstimate", p);

	if((p = json_strval(v, '.', "fields.timetracking.timespent")) != nil)
		if((q = json_strval(v, '.', "fields.timetracking.timespentSeconds")) != nil)
			fmtprint(&f, "%-16q %q %q\n", "timespent", p, q);
		else
			fmtprint(&f, "%-16q %q\n", "timespent", p);

	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(np, "timetracking", buf);
	free(buf);
}

static void
transitions(Node **n, Sess *s, char *issue)
{
	int i;
	Fmt f;
	Jvalue *v;
	char *buf, *name, *to;

	if((v = exch(s, nil, "rest/api/%s/issue/%U/transitions", s->api, issue)) == nil){
		warn("%s cannot get valid transition list, %r\n", issue);
		return;
	}

	fmtstrinit(&f);
	for(i = 0; (name = json_strval(v, '.', "transitions[%d].name", i)) != nil; i++){
		if((to = json_strval(v, '.', "transitions[%d].to.name", i)) == nil)
			continue;
		fmtprint(&f, "%-16q %q\n", name, to);
	}
	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	writehook(subtext(n, "transition", buf), add_transition, issue);
	free(buf);
	json_free(v);
}

static void
comments(Node *n, Jvalue *v, Sess *s)
{
	int i;
	long created, updated;
	char *p, *uid, *muid, *body, file[16];

	for(i = 0; (body = json_strval(v, '.', "fields.comment[%d].body", i)) != nil; i++){

		if((uid = json_strval(v, '.', "fields.comment[%d].author.name", i)) == nil)
			uid = "unknown";
		else
			adduser(s, uid);

		if((p = json_strval(v, '.', "fields.comment[%d].created", i)) != nil)
			created = iso2sec(p);
		else
			created = time(nil);

		if((p = json_strval(v, '.', "fields.comment[%d].updated", i)) != nil)
			updated = iso2sec(p);
		else
			updated = time(nil);

		if((p = json_strval(v, '.', "fields.comment[%d].updateAuthor.name", i)) != nil)
			muid = p;
		else
			muid = uid;

		snprint(file, sizeof(file), "%d", i);
		/* abuse mtime and atime for created and updated (ugh) */
		fulltext(&n->child, file, uid, nil, muid, created, updated, body);
	}
}


static void
dumptree(Node *n, int in)
{

	for (; n; n = n->next){
		print("%*s %D\n", in, "", &n->dir);
		if (n->child)
			dumptree(n->child, in+2);
	}
}

Node *
issueinfo_v5(Node *n, Sess *s, char *issue)
{
	Jvalue *v;
	Node *nn;
	char *p, *uid;

	if((v = exch(s, nil, "rest/api/%s/issue/%U", s->api, issue)) == nil){
		warn("%s cannot get issue detail, %r\n", issue);
		return n;
	}

	if((uid = json_strval(v, '.', "fields.reporter.name")) != nil){
		adduser(s, uid);
		fulltext(&n, "reporter", uid, nil, uid, 0, 0, uid);
	}
	else
		uid = "?";
	if((p = json_strval(v, '.', "fields.status.name")) != nil)
		fulltext(&n, "status", uid, nil, uid, 0, 0, p);
	if((p = json_strval(v, '.', "fields.resolution.name")) != nil)
		fulltext(&n, "resolution", uid, nil, uid, 0, 0, p);
	if((p = json_strval(v, '.', "fields.workratio")) != nil)
		fulltext(&n, "workratio", uid, nil, uid, 0, 0, p);
	if((p = json_strval(v, '.', "fields.votes.votes")) != nil)
		fulltext(&n, "votes", uid, nil, uid, 0, 0, p);
	if((p = json_strval(v, '.', "fields.priority.name")) != nil)
		fulltext(&n, "priority", uid, nil, uid, 0, 0, p);
	if((p = json_strval(v, '.', "fields.created")) != nil)
		fulltext(&n, "created", uid, nil, uid, 0, 0, ctime(iso2sec(p)));
	if((p = json_strval(v, '.', "fields.updated")) != nil)
		fulltext(&n, "updated", uid, nil, uid, 0, 0, ctime(iso2sec(p)));
	if((p = json_strval(v, '.', "fields.resolutiondate")) != nil)
		fulltext(&n, "resolved", uid, nil, uid, 0, 0, ctime(iso2sec(p)));
	if((p = json_strval(v, '.', "fields.environment")) != nil)
		fulltext(&n, "environment", uid, nil, uid, 0, 0, p);
	if((p = json_strval(v, '.', "FixVersions.name")) != nil)
		fulltext(&n, "fix-version", uid, nil, uid, 0, 0, p);

	if((p = json_strval(v, '.', "fields.assignee.name")) == nil)
		p = "";
	else
		adduser(s, p);
	writehook(fulltext(&n, "assignee", uid, nil, uid, 0, 0, p), upd_assignee, issue);

	if((p = json_strval(v, '.', "fields.issuetype.name")) == nil)
		p = "";
	writehook(fulltext(&n, "issuetype", uid, nil, uid, 0, 0, p), upd_summary, issue);

	if((p = json_strval(v, '.', "fields.summary")) == nil)
		p = "";
	writehook(fulltext(&n, "summary", uid, nil, uid, 0, 0, p), upd_summary, issue);

	if((p = json_strval(v, '.', "fields.description")) == nil)
		p = "";
	writehook(fulltext(&n, "description", uid, nil, uid, 0, 0, p), upd_description, issue);

	watchers(&n, v);
	fixversions(&n, v);
	components(&n, v);
	progress(&n, v, "progress.", "progress");
	progress(&n, v, "fields.aggregateprogress", "aggregateprogress");
	timetracking(&n, v);
	transitions(&n, s, issue);

	worklog(newdir(&n, nil, nil, "worklog"), v, s);
	attachments(newdir(&n, nil, nil, "attachments"), v, s);
	subtasks(newdir(&n, nil, nil, "subtasks"), v);
	links(newdir(&n, nil, nil, "links"), v);

	nn = newdir(&n, nil, nil, "comments");
	comments(nn, v, s);
	createhook(nn, new_comment, issue, 0);

	json_free(v);
	return n;
}
